from __future__ import print_function
import sys

if sys.version_info[0] < 3:
    print("pkuseg does not support python2", file=sys.stderr)
    sys.exit(1)

import os
import time
import pickle as pkl
import multiprocessing

from multiprocessing import Process, Queue

from . import trainer
from . import inference as _inf

from .config import config
from .feature_extractor import FeatureExtractor
from .model import Model
from .download import download_model
from .postag import Postag

class TrieNode:
    """建立词典的Trie树节点"""

    def __init__(self, isword):
        self.isword = isword
        self.usertag = ''
        self.children = {}


class Preprocesser:
    """预处理器，在用户词典中的词强制分割"""

    def __init__(self, dict_file):
        """初始化建立Trie树"""
        if dict_file is None:
            dict_file = []
        self.dict_data = dict_file
        if isinstance(dict_file, str):
            with open(dict_file, encoding="utf-8") as f:
                lines = f.readlines()
            self.trie = TrieNode(False)
            for line in lines:
                fields = line.strip().split('\t')
                word = fields[0].strip()
                usertag = fields[1].strip() if len(fields) > 1 else ''
                self.insert(word, usertag)
        else:
            self.trie = TrieNode(False)
            for w_t in dict_file:
                if isinstance(w_t, str):
                    w = w_t.strip()
                    t = ''
                else:
                    assert isinstance(w_t, tuple)
                    assert len(w_t)==2
                    w, t = map(lambda x:x.strip(), w_t)
                self.insert(w, t)

    def insert(self, word, usertag):
        """Trie树中插入单词"""
        l = len(word)
        now = self.trie
        for i in range(l):
            c = word[i]
            if not c in now.children:
                now.children[c] = TrieNode(False)
            now = now.children[c]
        now.isword = True
        now.usertag = usertag

    def solve(self, txt):
        """对文本进行预处理"""
        outlst = []
        iswlst = []
        taglst = []
        l = len(txt)
        last = 0
        i = 0
        while i < l:
            now = self.trie
            j = i
            found = False
            usertag = ''
            last_word_idx = -1 # 表示从当前位置i往后匹配，最长匹配词词尾的idx
            while True:
                c = txt[j]
                if not c in now.children and last_word_idx != -1:
                    found = True
                    break
                if not c in now.children and last_word_idx == -1:
                    break
                now = now.children[c]
                if now.isword:
                    last_word_idx = j
                    usertag = now.usertag
                j += 1
                if j == l and last_word_idx == -1:
                    break
                if j == l and last_word_idx != -1 :
                    j = last_word_idx + 1
                    found = True
                    break
            if found:
                if last != i:
                    outlst.append(txt[last:i])
                    iswlst.append(False)
                    taglst.append('')
                outlst.append(txt[i:j])
                iswlst.append(True)
                taglst.append(usertag)
                last = j
                i = j
            else:
                i += 1
        if last < l:
            outlst.append(txt[last:l])
            iswlst.append(False)
            taglst.append('')
        return outlst, iswlst, taglst

class Postprocesser:
    """对分词结果后处理"""
    def __init__(self, common_name, other_names):
        if common_name is None and other_names is None:
            self.do_process = False
            return
        self.do_process = True
        if common_name is None:
            self.common_words = set()
        else:
            # with open(common_name, encoding='utf-8') as f:
            #     lines = f.readlines()
            # self.common_words = set(map(lambda x:x.strip(), lines))
            with open(common_name, "rb") as f:
                all_words = pkl.load(f).strip().split("\n")
            self.common_words = set(all_words)
        if other_names is None:
            self.other_words = set()
        else:
            self.other_words = set()
            for other_name in other_names:
                # with open(other_name, encoding='utf-8') as f:
                #     lines = f.readlines()
                # self.other_words.update(set(map(lambda x:x.strip(), lines)))
                with open(other_name, "rb") as f:
                    all_words = pkl.load(f).strip().split("\n")
                self.other_words.update(set(all_words))

    def post_process(self, sent, check_seperated):
        for m in reversed(range(2, 8)): 
            end = len(sent)-m
            if end < 0:
                continue
            i = 0
            while (i < end + 1):
                merged_words = ''.join(sent[i:i+m])
                if merged_words in self.common_words:
                    do_seg = True
                elif merged_words in self.other_words:
                    if check_seperated:
                        seperated = all(((w in self.common_words) 
                            or (w in self.other_words)) for w in sent[i:i+m])
                    else:
                        seperated = False
                    if seperated:
                        do_seg = False
                    else:
                        do_seg = True
                else:
                    do_seg = False
                if do_seg:
                    for k in range(m):
                        del sent[i]
                    sent.insert(i, merged_words)
                    i += 1
                    end = len(sent) - m
                else:
                    i += 1 
        return sent

    def __call__(self, sent):
        if not self.do_process:
            return sent
        return self.post_process(sent, check_seperated=True)

class pkuseg:
    def __init__(self, model_name="spacy_ontonotes", user_dict="default", postag=False):
        """初始化函数，加载模型及用户词典"""
        # print("loading model")
        # config = Config()
        # self.config = config
        self.postag = postag
        if model_name in config.available_models:
            config.modelDir = os.path.join(
                config.pkuseg_home,
                model_name,
            )
            download_model(config.model_urls[model_name], config.pkuseg_home, config.model_hash[model_name])
        else:
            config.modelDir = model_name
        # config.fModel = os.path.join(config.modelDir, "model.txt")
        if user_dict is None:
            file_name = None
            other_names = None
        else:
            if user_dict not in config.available_models + ["default"]:
                file_name = user_dict
            else:
                file_name = None
            if model_name in config.models_with_dict:
                other_name = os.path.join(
                    config.pkuseg_home,
                    model_name,
                    model_name+"_dict.pkl",
                )
                default_name = os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "dicts", "default.pkl",
                )
                other_names = [other_name, default_name]
            else:
                default_name = os.path.join(
                    os.path.dirname(os.path.realpath(__file__)),
                    "dicts", "default.pkl",
                )
                other_names = [default_name]

        self.preprocesser = Preprocesser(file_name)
        # self.preprocesser = Preprocesser([])
        self.postprocesser = Postprocesser(None, other_names)

        self.feature_extractor = FeatureExtractor.load()
        self.model = Model.load()

        self.idx_to_tag = {
            idx: tag for tag, idx in self.feature_extractor.tag_to_idx.items()
        }

        self.n_feature = len(self.feature_extractor.feature_to_idx)
        self.n_tag = len(self.feature_extractor.tag_to_idx)

        if postag:
            download_model(config.model_urls["postag"], config.pkuseg_home, config.model_hash["postag"])
            postag_dir = os.path.join(
                config.pkuseg_home,
                "postag",
            )
            self.tagger = Postag(postag_dir)

        # print("finish")

    def _cut(self, text):
        """
        直接对文本分词
        """

        examples = list(self.feature_extractor.normalize_text(text))
        length = len(examples)

        all_feature = []  # type: List[List[int]]
        for idx in range(length):
            node_feature_idx = self.feature_extractor.get_node_features_idx(
                idx, examples
            )
            # node_feature = self.feature_extractor.get_node_features(
            #     idx, examples
            # )

            # node_feature_idx = []
            # for feature in node_feature:
            #     feature_idx = self.feature_extractor.feature_to_idx.get(feature)
            #     if feature_idx is not None:
            #         node_feature_idx.append(feature_idx)
            # if not node_feature_idx:
            #     node_feature_idx.append(0)

            all_feature.append(node_feature_idx)

        _, tags = _inf.decodeViterbi_fast(all_feature, self.model)

        words = []
        current_word = None
        is_start = True
        for tag, char in zip(tags, text):
            if is_start:
                current_word = char
                is_start = False
            elif "B" in self.idx_to_tag[tag]:
                words.append(current_word)
                current_word = char
            else:
                current_word += char
        if current_word:
            words.append(current_word)

        return words

    def cut(self, txt):
        """分词，结果返回一个list"""

        txt = txt.strip()

        ret = []
        usertags = []

        if not txt:
            return ret

        imary = txt.split()  # 根据空格分为多个片段

        # 对每个片段分词
        for w0 in imary:
            if not w0:
                continue

            # 根据用户词典拆成更多片段
            lst, isword, taglst = self.preprocesser.solve(w0)

            for w, isw, usertag in zip(lst, isword, taglst):
                if isw:
                    ret.append(w)
                    usertags.append(usertag)
                    continue

                output = self._cut(w)
                post_output = self.postprocesser(output)
                ret.extend(post_output)
                usertags.extend(['']*len(post_output))
        
        if self.postag:
            tags = self.tagger.tag(ret.copy())
            for i, usertag in enumerate(usertags):
                if usertag:
                    tags[i] = usertag
            ret = list(zip(ret, tags))
        return ret


def train(trainFile, testFile, savedir, train_iter=20, init_model=None):
    """用于训练模型"""
    # config = Config()
    starttime = time.time()
    if not os.path.exists(trainFile):
        raise Exception("trainfile does not exist.")
    if not os.path.exists(testFile):
        raise Exception("testfile does not exist.")
    if not os.path.exists(config.tempFile):
        os.makedirs(config.tempFile)
    if not os.path.exists(config.tempFile + "/output"):
        os.mkdir(config.tempFile + "/output")
    # config.runMode = "train"
    config.trainFile = trainFile
    config.testFile = testFile
    config.modelDir = savedir
    # config.fModel = os.path.join(config.modelDir, "model.txt")
    config.nThread = 1
    config.ttlIter = train_iter
    config.init_model = init_model

    os.makedirs(config.modelDir, exist_ok=True)

    trainer.train(config)

    # pkuseg.main.run(config)
    # clearDir(config.tempFile)
    print("Total time: " + str(time.time() - starttime))


def _test_single_proc(
    input_file, output_file, model_name="default", user_dict="default", postag=False, verbose=False
):

    times = []
    times.append(time.time())
    seg = pkuseg(model_name, user_dict, postag=postag)

    times.append(time.time())
    if not os.path.exists(input_file):
        raise Exception("input_file {} does not exist.".format(input_file))
    with open(input_file, "r", encoding="utf-8") as f:
        lines = f.readlines()

    times.append(time.time())
    results = []
    for line in lines:
        if not postag:
            results.append(" ".join(seg.cut(line)))
        else:
            results.append(" ".join(map(lambda x:"/".join(x), seg.cut(line))))

    times.append(time.time())
    with open(output_file, "w", encoding="utf-8") as f:
        f.write("\n".join(results))
    times.append(time.time())

    print("total_time:\t{:.3f}".format(times[-1] - times[0]))

    if verbose:
        time_strs = ["load_model", "read_file", "word_seg", "write_file"]
        for key, value in zip(
            time_strs,
            [end - start for start, end in zip(times[:-1], times[1:])],
        ):
            print("{}:\t{:.3f}".format(key, value))


def _proc_deprecated(seg, lines, start, end, q):
    for i in range(start, end):
        l = lines[i].strip()
        ret = seg.cut(l)
        q.put((i, " ".join(ret)))


def _proc(seg, in_queue, out_queue):
    # TODO: load seg (json or pickle serialization) in sub_process
    #       to avoid pickle seg online when using start method other
    #       than fork
    while True:
        item = in_queue.get()
        if item is None:
            return
        idx, line = item
        if not seg.postag:
            output_str = " ".join(seg.cut(line))
        else:
            output_str = " ".join(map(lambda x:"/".join(x), seg.cut(line)))
        out_queue.put((idx, output_str))


def _proc_alt(model_name, user_dict, postag, in_queue, out_queue):
    seg = pkuseg(model_name, user_dict, postag=postag)
    while True:
        item = in_queue.get()
        if item is None:
            return
        idx, line = item
        if not postag:
            output_str = " ".join(seg.cut(line))
        else:
            output_str = " ".join(map(lambda x:"/".join(x), seg.cut(line)))
        out_queue.put((idx, output_str))


def _test_multi_proc(
    input_file,
    output_file,
    nthread,
    model_name="default",
    user_dict="default",
    postag=False,
    verbose=False,
):

    alt = multiprocessing.get_start_method() == "spawn"

    times = []
    times.append(time.time())

    if alt:
        seg = None
    else:
        seg = pkuseg(model_name, user_dict, postag)

    times.append(time.time())
    if not os.path.exists(input_file):
        raise Exception("input_file {} does not exist.".format(input_file))
    with open(input_file, "r", encoding="utf-8") as f:
        lines = f.readlines()

    times.append(time.time())
    in_queue = Queue()
    out_queue = Queue()
    procs = []
    for _ in range(nthread):
        if alt:
            p = Process(
                target=_proc_alt,
                args=(model_name, user_dict, postag, in_queue, out_queue),
            )
        else:
            p = Process(target=_proc, args=(seg, in_queue, out_queue))
        procs.append(p)

    for idx, line in enumerate(lines):
        in_queue.put((idx, line))

    for proc in procs:
        in_queue.put(None)
        proc.start()

    times.append(time.time())
    result = [None] * len(lines)
    for _ in result:
        idx, line = out_queue.get()
        result[idx] = line

    times.append(time.time())
    for p in procs:
        p.join()

    times.append(time.time())
    with open(output_file, "w", encoding="utf-8") as f:
        f.write("\n".join(result))
    times.append(time.time())

    print("total_time:\t{:.3f}".format(times[-1] - times[0]))

    if verbose:
        time_strs = [
            "load_model",
            "read_file",
            "start_proc",
            "word_seg",
            "join_proc",
            "write_file",
        ]

        if alt:
            times = times[1:]
            time_strs = time_strs[1:]
            time_strs[2] = "load_modal & word_seg"

        for key, value in zip(
            time_strs,
            [end - start for start, end in zip(times[:-1], times[1:])],
        ):
            print("{}:\t{:.3f}".format(key, value))


def test(
    input_file,
    output_file,
    model_name="default",
    user_dict="default",
    nthread=10,
    postag=False,
    verbose=False,
):

    if nthread > 1:
        _test_multi_proc(
            input_file, output_file, nthread, model_name, user_dict, postag, verbose
        )
    else:
        _test_single_proc(
            input_file, output_file, model_name, user_dict, postag, verbose
        )




# from .config import Config
# from pkuseg.feature_generator import
# import os
import copy
import random


# class dataFormat:
#     def __init__(self, config):
#         self.featureIndexMap = {}
#         self.tagIndexMap = {}
#         self.config = config

#     def convert(self):
#         config = self.config
#         if config.runMode.find("train") >= 0:
#             self.getMaps(config.fTrain)
#             self.saveFeature(config.modelDir + "/featureIndex.txt")
#             self.convertFile(config.fTrain)
#         else:
#             self.readFeature(config.modelDir + "/featureIndex.txt")
#             self.readTag(config.modelDir + "/tagIndex.txt")
#         self.convertFile(config.fTest)
#         if config.dev:
#             self.convertFile(config.fDev)

#     def saveFeature(self, file):
#         featureList = list(self.featureIndexMap.keys())
#         num = len(featureList) // 10
#         for i in range(10):
#             l = i * num
#             r = (i + 1) * num if i < 9 else len(featureList)
#             with open(file + "_" + str(i), "w", encoding="utf-8") as sw:
#                 for w in range(l, r):
#                     word = featureList[w]
#                     sw.write(word + " " + str(self.featureIndexMap[word]) + "\n")

#     def readFeature(self, file):
#         featureList = []
#         for i in range(10):
#             featureList.append([])
#             with open(file + "_" + str(i), encoding="utf-8") as f:
#                 lines = f.readlines()
#             for line in lines:
#                 featureList[i].append(line.strip())
#         feature = []
#         for i in range(10):
#             for line in featureList[i]:
#                 word, index = line.split(" ")
#                 self.featureIndexMap[word] = int(index)

#     def readFeatureNormal(self, path):
#         with open(path, encoding="utf-8") as f:
#             lines = f.readlines()
#         for line in lines:
#             u, v = line.split(" ")
#             self.featureIndexMap[u] = int(v)

#     def readTag(self, path):
#         with open(path, encoding="utf-8") as f:
#             lines = f.readlines()
#         for line in lines:
#             u, v = line.split(" ")
#             self.tagIndexMap[u] = int(v)

#     def getMaps(self, file):
#         config = self.config
#         if not os.path.exists(file):
#             print("file {} not exist!".format(file))
#         print("file {} converting...".format(file))
#         featureFreqMap = {}
#         tagSet = set()
#         with open(file, encoding="utf-8") as f:
#             lines = f.readlines()
#         for line in lines:
#             line = line.replace("\t", " ")
#             line = line.replace("\r", "").strip()
#             if line == "":
#                 continue
#             ary = line.split(config.blank)
#             for i in range(1, len(ary) - 1):
#                 if ary[i] == "" or ary[i] == "/":
#                     continue
#                 if config.weightRegMode == "GL":
#                     if not config.GL_init and config.groupTrim[i - 1]:
#                         continue

#                 ary2 = ary[i].split(config.slash)
#                 feature = str(i) + "." + ary2[0]
#                 if not feature in featureFreqMap:
#                     featureFreqMap[feature] = 0
#                 featureFreqMap[feature] += 1
#             tag = ary[-1]
#             tagSet.add(tag)
#         sortList = []
#         for k in featureFreqMap:
#             sortList.append(k + " " + str(featureFreqMap[k]))
#         if config.weightRegMode == "GL":
#             sortList.sort(key=lambda x: (int(x.split(config.blank)[1].strip()), x))
#             with open("featureTemp_sorted.txt", "w", encoding="utf-8") as f:
#                 for x in sortList:
#                     f.write(x + "\n")
#             config.groupStart = [0]
#             config.groupEnd = []
#             for k in range(1, len(sortList)):
#                 thisAry = sortList[k].split(config.dot)
#                 preAry = sortList[k - 1].split(config.dot)
#                 s = thisAry[0]
#                 preAry = preAry[0]
#                 if s != preAry:
#                     config.groupStart.append(k)
#                     config.groupEnd.append(k)
#             config.groupEnd.append(len(sortList))
#         else:
#             sortList.sort(
#                 key=lambda x: (int(x.split(config.blank)[1].strip()), x), reverse=True
#             )

#         if config.weightRegMode == "GL" and config.GL_init:
#             if nFeatTemp != len(config.groupStart):
#                 raise Exception(
#                     "inconsistent # of features per line, check the feature file for consistency!"
#                 )
#         with open(
#             os.path.join(config.modelDir, "featureIndex.txt"), "w", encoding="utf-8"
#         ) as swFeat:
#             for i, l in enumerate(sortList):
#                 ary = l.split(config.blank)
#                 self.featureIndexMap[ary[0]] = i
#                 swFeat.write("{} {}\n".format(ary[0].strip(), i))
#         with open(os.path.join(config.modelDir, "tagIndex.txt"), "w", encoding="utf-8") as swTag:
#             tagSortList = []
#             for tag in tagSet:
#                 tagSortList.append(tag)
#             tagSortList.sort()
#             for i, l in enumerate(tagSortList):
#                 self.tagIndexMap[l] = i
#                 swTag.write("{} {}\n".format(l, i))

#     def convertFile(self, file):
#         config = self.config
#         if not os.path.exists(file):
#             print("file {} not exist!".format(file))
#         print("file converting...")
#         if file == config.fTrain:
#             swFeature = open(config.fFeatureTrain, "w", encoding="utf-8")
#             swGold = open(config.fGoldTrain, "w", encoding="utf-8")
#         else:
#             swFeature = open(config.fFeatureTest, "w", encoding="utf-8")
#             swGold = open(config.fGoldTest, "w", encoding="utf-8")
#         swFeature.write(str(len(self.featureIndexMap)) + "\n\n")
#         swGold.write(str(len(self.tagIndexMap)) + "\n\n")
#         with open(file, encoding="utf-8") as sr:
#             readLines = sr.readlines()
#         featureList = []
#         goldList = []
#         for k in range(len(readLines)):
#             line = readLines[k]
#             line = line.replace("\t", "").strip()
#             featureLine = ""
#             goldLine = ""
#             if line == "":
#                 featureLine = featureLine + "\n"
#                 goldLine = goldLine + "\n\n"
#                 featureList.append(featureLine)
#                 goldList.append(goldLine)
#                 continue
#             flag = 0
#             ary = line.split(config.blank)
#             tmp = []
#             for i in ary:
#                 if i != "":
#                     tmp.append(i)
#             ary = tmp
#             for i in range(1, len(ary) - 1):
#                 if ary[i] == "/":
#                     continue
#                 ary2 = ary[i].split(config.slash)
#                 tmp = []
#                 for j in ary2:
#                     if j != "":
#                         tmp.append(j)
#                 ary2 = tmp
#                 feature = str(i) + "." + ary2[0]
#                 value = ""
#                 real = False
#                 if len(ary2) > 1:
#                     value = ary2[1]
#                     real = True
#                 if not feature in self.featureIndexMap:
#                     continue
#                 flag = 1
#                 fIndex = self.featureIndexMap[feature]
#                 if not real:
#                     featureLine = featureLine + str(fIndex) + ","
#                 else:
#                     featureLine = featureLine + str(fIndex) + "/" + value + ","
#             if flag == 0:
#                 featureLine = featureLine + "0"
#             featureLine = featureLine + "\n"
#             tag = ary[-1]
#             tIndex = self.tagIndexMap[tag]
#             goldLine = goldLine + str(tIndex) + ","
#             featureList.append(featureLine)
#             goldList.append(goldLine)
#         for i in range(len(featureList)):
#             swFeature.write(featureList[i])
#             swGold.write(goldList[i])
#         swFeature.close()
#         swGold.close()


class DataSet:
    def __init__(self, n_tag=0, n_feature=0):
        self.lst = []  # type: List[Example]
        self.n_tag = n_tag
        self.n_feature = n_feature
        # if len(args) == 2:
        #     if type(args[0]) == int:
        #         self.nTag, self.nFeature = args
        #     else:
        #         self.load(args[0], args[1])

    def __len__(self):
        return len(self.lst)

    def __iter__(self):
        return self.iterator()

    def __getitem__(self, x):
        return self.lst[x]

    def iterator(self):
        for i in self.lst:
            yield i

    def append(self, x):
        self.lst.append(x)

    def clear(self):
        self.lst = []

    def randomShuffle(self):
        cp = copy.deepcopy(self)
        random.shuffle(cp.lst)
        return cp

    # def setDataInfo(self, X):
    #     self.nTag = X.nTag
    #     self.nFeature = X.nFeature

    def resize(self, scale):
        dataset = DataSet(self.n_tag, self.n_feature)
        new_size = int(len(self) * scale)
        old_size = len(self)
        for i in range(new_size):
            if i >= old_size:
                i %= old_size
            dataset.append(self[i])
        return dataset

    @classmethod
    def load(cls, feature_idx_file, tag_idx_file):
        dataset = cls.__new__(cls)

        # def load(self, fileFeature, fileTag):
        with open(feature_idx_file, encoding="utf-8") as f_reader, open(
            tag_idx_file, encoding="utf-8"
        ) as t_reader:

            example_strs = f_reader.read().split("\n\n")[:-1]
            tags_strs = t_reader.read().split("\n\n")[:-1]

        assert len(example_strs) == len(
            tags_strs
        ), "lengths do not match:\t{}\n{}\n".format(example_strs, tags_strs)

        n_feature = int(example_strs[0])
        n_tag = int(tags_strs[0])

        dataset.n_feature = n_feature
        dataset.n_tag = n_tag
        dataset.lst = []

        for example_str, tags_str in zip(example_strs[1:], tags_strs[1:]):
            features = [
                list(map(int, feature_line.split(",")))
                for feature_line in example_str.split("\n")
            ]
            tags = tags_str.split(",")
            example = Example(features, tags)
            dataset.lst.append(example)

        return dataset
        # txt = srfileFeature.read()
        # txt.replace("\r", "")
        # fAry = txt.split(Config.biLineEnd)
        # tmp = []
        # for i in fAry:
        #     if i != "":
        #         tmp.append(i)
        # fAry = tmp
        # txt = srfileTag.read()
        # txt.replace("\r", "")
        # tAry = txt.split(Config.biLineEnd)
        # tmp = []
        # for i in tAry:
        #     if i != "":
        #         tmp.append(i)
        # tAry = tmp

        # assert len(fAry) == len(tAry)
        # self.nFeature = int(fAry[0])
        # self.nTag = int(tAry[0])
        # for i in range(1, len(fAry)):
        #     features = fAry[i]
        #     tags = tAry[i]
        #     seq = dataSeq()
        #     seq.read(features, tags)
        #     self.append(seq)

    # @property
    # def NTag(self):
    #     return self.nTag


class Example:
    def __init__(self, features, tags):
        self.features = features  # type: List[List[int]]
        self.tags = list(map(int, tags))  # type: List[int]
        self.predicted_tags = None

    def __len__(self):
        return len(self.features)


# class dataSeq:
#     def __init__(self, *args):
#         self.featureTemps = []
#         self.yGold = []
#         if len(args) == 2:
#             self.featureTemps = copy.deepcopy(args[0])
#             self.yGold = copy.deepcopy(args[1])
#         elif len(args) == 3:
#             x, n, length = args
#             end = min(n + length, len(x))
#             for i in range(n, end):
#                 self.featureTemps.append(x.featureTemps[i])
#                 yGold.append(x.yGold[i])

#     def __len__(self):
#         return len(self.featureTemps)

#     def read(self, a, b):
#         lineAry = a.split(Config.lineEnd)
#         for im in lineAry:
#             if im == "":
#                 continue
#             nodeList = []
#             imAry = im.split(Config.comma)
#             for imm in imAry:
#                 if imm == "":
#                     continue
#                 if imm.find("/") >= 0:
#                     biAry = imm.split(Config.slash)
#                     ft = featureTemp(int(biAry[0], float(biAry[1])))
#                     nodeList.append(ft)
#                 else:
#                     ft = featureTemp(int(imm), 1)
#                     nodeList.append(ft)
#             self.featureTemps.append(nodeList)
#         lineAry = b.split(Config.comma)
#         for im in lineAry:
#             if im == "":
#                 continue
#             self.yGold.append(int(im))

#     # def load(self, feature):
#     #     for imAry in feature:
#     #         nodeList = []
#     #         for imm in imAry:
#     #             if imm == "":
#     #                 continue
#     #             if imm.find("/") >= 0:
#     #                 biAry = imm.split(Config.slash)
#     #                 ft = featureTemp(int(biAry[0], float(biAry[1])))
#     #                 nodeList.append(ft)
#     #             else:
#     #                 ft = featureTemp(int(imm), 1)
#     #                 nodeList.append(ft)
#     #         self.featureTemps.append(nodeList)
#     #         self.yGold.append(0)

#     def getFeatureTemp(self, *args):
#         return (
#             self.featureTemps if len(args) == 0 else self.featureTemps[args[0]]
#         )

#     def getTags(self, *args):
#         return self.yGold if len(args) == 0 else self.yGold[args[0]]

#     def setTags(self, lst):
#         assert len(lst) == len(self.yGold)
#         for i in range(len(lst)):
#             self.yGold[i] = lst[i]


# class dataSeqTest:
#     def __init__(self, x, yOutput):
#         self._x = x
#         self._yOutput = yOutput



import os
import sys
import numpy as np

from .config import config


class Model:
    def __init__(self, n_feature, n_tag):

        self.n_tag = n_tag
        self.n_feature = n_feature
        self.n_transition_feature = n_tag * (n_feature + n_tag)
        if config.random:
            self.w = np.random.random(size=(self.n_transition_feature,)) * 2 - 1
        else:
            self.w = np.zeros(self.n_transition_feature)

    def expand(self, n_feature, n_tag):
        new_transition_feature = n_tag * (n_feature + n_tag)
        if config.random:
            new_w = np.random.random(size=(new_transition_feature,)) * 2 - 1
        else:
            new_w = np.zeros(new_transition_feature)
        n_node = self.n_tag * self.n_feature
        n_edge = self.n_tag * self.n_tag
        new_w[:n_node] = self.w[:n_node]
        new_w[-n_edge:] = self.w[-n_edge:]
        self.n_tag = n_tag
        self.n_feature = n_feature
        self.n_transition_feature = new_transition_feature
        self.w = new_w

    def _get_node_tag_feature_id(self, feature_id, tag_id):
        return feature_id * self.n_tag + tag_id

    def _get_tag_tag_feature_id(self, pre_tag_id, tag_id):
        return self.n_feature * self.n_tag + tag_id * self.n_tag + pre_tag_id

    @classmethod
    def load(cls, model_dir=None):
        if model_dir is None:
            model_dir = config.modelDir
        model_path = os.path.join(model_dir, "weights.npz")
        if os.path.exists(model_path):
            npz = np.load(model_path)
            sizes = npz["sizes"]
            w = npz["w"]
            model = cls.__new__(cls)
            model.n_tag = int(sizes[0])
            model.n_feature = int(sizes[1])
            model.n_transition_feature = model.n_tag * (
                model.n_feature + model.n_tag
            )
            model.w = w
            assert model.w.shape[0] == model.n_transition_feature
            return model

        print(
            "WARNING: weights.npz does not exist, try loading using old format",
            file=sys.stderr,
        )

        model_path = os.path.join(model_dir, "model.txt")
        with open(model_path, encoding="utf-8") as f:
            ary = f.readlines()

        model = cls.__new__(cls)
        model.n_tag = int(ary[0].strip())
        wsize = int(ary[1].strip())
        w = np.zeros(wsize)
        for i in range(2, wsize):
            w[i - 2] = float(ary[i].strip())
        model.w = w
        model.n_feature = wsize // model.n_tag - model.n_tag
        model.n_transition_feature = wsize

        model.save(model_dir)
        return model

    @classmethod
    def new(cls, model, copy_weight=True):

        new_model = cls.__new__(cls)
        new_model.n_tag = model.n_tag
        if copy_weight:
            new_model.w = model.w.copy()
        else:
            new_model.w = np.zeros_like(model.w)
        new_model.n_feature = (
            new_model.w.shape[0] // new_model.n_tag - new_model.n_tag
        )
        new_model.n_transition_feature = new_model.w.shape[0]
        return new_model

    def save(self, model_dir=None):
        if model_dir is None:
            model_dir = config.modelDir
        sizes = np.array([self.n_tag, self.n_feature])
        np.savez(
            os.path.join(model_dir, "weights.npz"), sizes=sizes, w=self.w
        )
        # np.save
        # with open(file, "w", encoding="utf-8") as f:
        #     f.write("{}\n{}\n".format(self.n_tag, self.w.shape[0]))
        #     for value in self.w:
        #         f.write("{:.4f}\n".format(value))
